from yta_validation.parameter import ParameterValidator
from yta_programming.decorators.requires_dependency import requires_dependency
from typing import Union
from abc import ABC, abstractmethod


class _ImageDescriptor(ABC):
    """
    Abstract class to describe images.
    """

    @abstractmethod
    def describe(
        self,
        image: Union[str, 'Image.Image', 'np.ndarray']
    ) -> str:
        """
        Describe the provided 'image' using an engine
        capable of it.
        """
        pass

class DefaultImageDescriptor(_ImageDescriptor):
    """
    Default class to describe an image. It will choose the
    engine we think is a good choice in general.

    The process could take from seconds to a couple of minutes
    according to the system specifications.
    """

    def describe(
        self,
        image: Union[str, 'Image.Image', 'np.ndarray']
    ) -> str:
        ParameterValidator.validate_mandatory_instance_of('image', image, [str, 'Image.Image', 'np.ndarray'])

        return BlipImageDescriptor().describe(image)

class BlipImageDescriptor(_ImageDescriptor):
    """
    Class to describe an image using the Blip engine, which
    is from Salesforce and will use pretrained models that are
    stored locally in 'C:/Users/USER/.cache/huggingface/hub',
    loaded in memory and used to describe it.

    The process could take from seconds to a couple of minutes
    according to the system specifications.

    This class needs this library installed:
    - `yta_ai_blip`
    """

    @requires_dependency('yta_ai_blip', 'yta_image_advanced_ai', 'yta_ai_blip')
    def describe(
        self,
        image: Union[str, 'Image.Image', 'np.ndarray']
    ) -> str:
        from yta_ai_blip import Blip

        ParameterValidator.validate_mandatory_instance_of('image', image, [str, 'Image.Image', 'np.ndarray'])

        return Blip.describe(image)
    
class LlavaImageDescriptor(_ImageDescriptor):
    """
    Class to describe an image using the Llava engine
    through the 'ollama' python package.

    This class needs this library installed:
    - `yta_ai_llava`
    """

    @requires_dependency('yta_ai_llava', 'yta_image_advanced_ai', 'yta_ai_llava')
    def describe(
        self,
        image_filename: str
    ):
        """
        THIS METHOD IS NOT WORKING YET.

        TODO: This is not working because of my pc limitations.
        It cannot load the resources due to memory capacity.
        """
        from yta_ai_llava import Llava

        ParameterValidator.validate_mandatory_string('image_filename', image_filename, do_accept_empty = False)

        return Llava.describe(image_filename)


"""
AI Image generation file that contains the classes
capable to generate AI images. These classes will
raise Exceptions if the parameters are not provided
and will use internal functionality to do it.

Programmer help: The classes implement parameter
validation and raise exceptions, but the other files
from which other methods are imported do not 
implement them, so make sure you pass the right
and expected parameters. This should be the ideal
structure to keep in code, but you know... I write
code very fast so I can't go back and review and
refactor code often... So sorry about that :P
"""
from yta_general.dataclasses import FileReturned
from yta_constants.file import FileType
from yta_programming.output import Output
from yta_programming.decorators.requires_dependency import requires_dependency
from yta_validation.parameter import ParameterValidator
from abc import ABC, abstractmethod
from typing import Union


class _AIImageGenerator(ABC):
    """
    Abstract class to be inherited by any specific
    AI image generator.
    """

    @abstractmethod
    def generate_image(
        self,
        prompt: str,
        output_filename: Union[str, None] = None
    ) -> FileReturned:
        """
        Generate an image with the given 'prompt' and
        store it locally if 'output_filename' is 
        provided.
        """
        pass

class DefaultImageGenerator(_AIImageGenerator):
    """
    Default AI image generator. Useful when you don't
    know which engine you should use. We have choosen
    one that is a good choice for a general context.
    """

    def generate_image(
        self,
        prompt: str,
        output_filename: Union[str, None] = None
    ) -> FileReturned:
        """
        Generate an AI image and return 2 values, the
        image read with pillow and the final output
        filename used to store the image locally.
        """
        return ProdiaAIImageGenerator().generate_image(prompt, output_filename)

class ProdiaAIImageGenerator(_AIImageGenerator):
    """
    Prodia AI image generator.

    This class needs this library installed:
    - `yta_ai_prodia`
    """

    @requires_dependency('yta_ai_prodia', 'yta_image_advanced_ai', 'yta_ai_prodia')
    def generate_image(
        self,
        prompt: str,
        output_filename: Union[str, None] = None
    ) -> FileReturned:
        """
        Generate an AI image and return 2 values, the
        image read with pillow and the final output
        filename used to store the image locally.
        """
        from yta_ai_prodia import Prodia

        ParameterValidator.validate_mandatory_string('prompt', prompt, do_accept_empty = False)

        output_filename = Output.get_filename(output_filename, FileType.IMAGE)
        
        return Prodia.generate_image(prompt, output_filename)
    
class PollinationsAIImageGenerator(_AIImageGenerator):
    """
    Pollinations AI image generator.

    This is using the Pollinations platform wich
    contains an AI image generator API and 
    open-source model.

    Source: https://pollinations.ai/

    This class needs this library installed:
    - `yta_ai_pollinations`
    """

    @requires_dependency('yta_ai_pollinations', 'yta_image_advanced_ai', 'yta_ai_pollinations')
    def generate_image(
        self,
        prompt: str,
        output_filename: Union[str, None] = None
    ) -> FileReturned:
        """
        Generate an image with the Pollinations AI image
        generation model using the provided 'prompt' and
        stores it locally as 'output_filename'.
        """
        from yta_ai_pollinations import Pollinations

        ParameterValidator.validate_mandatory_string('prompt', prompt, do_accept_empty = False)

        output_filename = Output.get_filename(output_filename, FileType.IMAGE)
        
        return Pollinations.generate_image(prompt, output_filename)


"""
This module has been programmed to need the
'yta_image_advanced' only if you are using the
'Image' class that this file contains.
"""
from yta_image_advanced_ai.descriptor import DefaultImageDescriptor
from yta_programming.decorators.requires_dependency import requires_dependency


class Image:
    """
    TODO: Do I need to put the description here
    instead of in the '_Image' instance I return (?)
    """

    @requires_dependency('yta_image_advanced', 'yta_image_advanced_ai', 'yta_image_advanced')
    def __new__(
        cls,
        *args,
        **kwargs
    ):
        try:
            from yta_image_advanced import Image as AdvancedImage
        except ImportError as e:
            raise ImportError(
                f'The "Image" class of this module needs the "yta_image_advanced" module installed.'
            ) from e

        class _Image(AdvancedImage):
            """
            Advanced Image class that includes AI-related
            functionality.
            """

            @property
            def description(
                self
            ) -> str:
                """
                A description of the image, given by an engine that
                has been trained to describe images.
                """
                if not hasattr(self, '_description'):
                    self._description = DefaultImageDescriptor().describe(self.image)

                return self._description
            
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
        
        instance = _Image(*args, **kwargs)

        return instance